# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import os
from abc import ABC
from typing import Union

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.optimizers import Optimizer
from tensorflow.python.keras.callbacks import (
    BaseLogger,
    Callback,
    CSVLogger,
    ModelCheckpoint,
    ProgbarLogger,
    TensorBoard,
)

import mlrun

from .._common import MLRunInterface
from .callbacks import LoggingCallback
from .utils import TFKerasTypes


class TFKerasMLRunInterface(MLRunInterface, ABC):
    """
    Interface for adding MLRun features for tensorflow keras API.
    """

    # MLRun's context default name:
    DEFAULT_CONTEXT_NAME = "mlrun-tf-keras"

    # Attributes to be inserted so the MLRun interface will be fully enabled.
    _PROPERTIES = {
        # Logging callbacks list:
        "_logging_callbacks": set(),  # type: Set[Callback]
        # Variable to hold the horovod module:
        "_hvd": None,  # type: ModuleType
        # List of all the callbacks that should only be applied on rank 0 when using horovod:
        "_RANK_0_ONLY_CALLBACKS": {  # type: Set[str]
            "LoggingCallback",
            "MLRunLoggingCallback",
            "TensorboardLoggingCallback",
            ModelCheckpoint.__name__,
            TensorBoard.__name__,
            ProgbarLogger.__name__,
            CSVLogger.__name__,
            BaseLogger.__name__,
        },
    }
    _METHODS = [
        "add_logging_callback",
        "use_horovod",
        "note_rank_0_callback",
        "_pre_compile",
        "_pre_fit",
        "_pre_evaluate",
    ]

    # Attributes to replace so the MLRun interface will be fully enabled.
    _REPLACED_METHODS = ["compile", "fit", "evaluate"]

    @classmethod
    def add_interface(
        cls,
        obj: keras.Model,
        restoration: TFKerasTypes.MLRunInterfaceRestorationType = None,
    ):
        """
        Enrich the object with this interface properties, methods and functions, so it will have this TensorFlow.Keras
        MLRun's features.

        :param obj:                     The object to enrich his interface.
        :param restoration: Restoration information tuple as returned from 'remove_interface' in order to
                                        add the interface in a certain state.
        """
        super().add_interface(obj=obj, restoration=restoration)

    def mlrun_compile(self, *args, **kwargs):
        """
        MLRun's tf.keras.Model.compile wrapper. It will setup the optimizer when using horovod. The optimizer must be
        passed in a keyword argument and when using horovod, it must be passed as an Optimizer instance, not a string.

        :raise MLRunInvalidArgumentError: In case the optimizer provided did not follow the instructions above.
        """
        # Validate the optimizer is passed via keyword:
        if "optimizer" not in kwargs:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "The optimizer must be passed as a keyword argument:\n"
                "model.compile(\n"
                "    optimizer=...\n"
                ")"
            )

        # Call the pre compile method:
        optimizer = self._pre_compile(optimizer=kwargs["optimizer"])

        # Assign parameters:
        kwargs["optimizer"] = optimizer

        # Call the original compile method:
        return self.original_compile(*args, **kwargs)

    @classmethod
    def mlrun_fit(cls):
        """
        MLRun's tf.keras.Model.fit wrapper. It will setup the optimizer when using horovod. The optimizer must be
        passed in a keyword argument and when using horovod, it must be passed as an Optimizer instance, not a string.

        :raise MLRunInvalidArgumentError: In case the optimizer provided did not follow the instructions above.
        """

        def wrapper(self: keras.Model, *args, **kwargs):
            # Restore the evaluation method as fit will use it:
            cls._restore_attribute(obj=self, attribute_name="evaluate")

            # Setup the callbacks list:
            if "callbacks" not in kwargs or kwargs["callbacks"] is None:
                kwargs["callbacks"] = []

            # Add auto logging callbacks if they were added:
            kwargs["callbacks"] = kwargs["callbacks"] + list(self._logging_callbacks)

            # Setup default values if needed:
            kwargs["verbose"] = kwargs.get("verbose", 1)
            kwargs["steps_per_epoch"] = kwargs.get("steps_per_epoch", None)
            kwargs["validation_steps"] = kwargs.get("validation_steps", None)
            kwargs["validation_data"] = kwargs.get("validation_data", None)

            # Call the pre fit method:
            (
                callbacks,
                verbose,
                steps_per_epoch,
                validation_steps,
            ) = self._pre_fit(
                callbacks=kwargs["callbacks"],
                verbose=kwargs["verbose"],
                steps_per_epoch=kwargs["steps_per_epoch"],
                validation_steps=kwargs["validation_steps"],
            )

            # Assign parameters:
            kwargs["callbacks"] = callbacks
            kwargs["verbose"] = verbose
            kwargs["steps_per_epoch"] = steps_per_epoch
            kwargs["validation_steps"] = validation_steps

            # Call the original fit method:
            result = self.original_fit(*args, **kwargs)

            # Replace the evaluation method again:
            cls._replace_function(obj=self, function_name="evaluate")

            return result

        return wrapper

    def mlrun_evaluate(self, *args, **kwargs):
        """
        MLRun tf.keras.Model.evaluate wrapper. Will enable automatic logging if set.
        """
        # Setup the callbacks list:
        if "callbacks" not in kwargs or kwargs["callbacks"] is None:
            kwargs["callbacks"] = []

        # Add auto log callbacks if they were added:
        kwargs["callbacks"] = kwargs["callbacks"] + list(self._logging_callbacks)

        # Setup default values if needed:
        kwargs["steps"] = kwargs.get("steps", None)

        # Call the pre evaluate method:
        (callbacks, steps) = self._pre_evaluate(
            callbacks=kwargs["callbacks"],
            steps=kwargs["steps"],
        )

        # Assign parameters:
        kwargs["callbacks"] = callbacks
        kwargs["steps"] = steps

        # Call the original evaluation method:
        return self.original_evaluate(*args, **kwargs)

    def add_logging_callback(self, logging_callback: LoggingCallback):
        """
        Add the given logging callback to model's logging callbacks list. For further information regarding the logging
        callbacks, see 'mlrun.frameworks.tf_keras.callbacks.MLRunLoggingCallback' and
        'mlrun.frameworks.tf_keras.callbacks.TensorboardLoggingCallback'.

        :param logging_callback: The logging callback to add.
        """
        # If horovod is being used, there is no need to add the logging callbacks to ranks other than 0:
        if self._hvd is not None and self._hvd.rank() != 0:
            return

        # Add the logging callback:
        self._logging_callbacks.add(logging_callback)

    # TODO: Add horovod callbacks configurations. If not set (None), use the defaults.
    def use_horovod(self):
        """
        Setup the model or wrapped model to run with horovod.
        """
        # Import horovod:
        self._hvd = importlib.import_module("horovod.tensorflow.keras")

        # Initialize horovod:
        self._hvd.init()

    def note_rank_0_callback(self, callback_name: str):
        """
        Note an additional custom callback to be applied only on rank 0 when using horovod.

        :param callback_name: The name of the callback.
        """
        self._RANK_0_ONLY_CALLBACKS.add(callback_name)

    def _pre_compile(self, optimizer: Optimizer) -> Optimizer:
        """
        Method to call before calling 'compile' to setup the run and inputs for using horovod.

        :param optimizer: The optimzier to compile. It will be wrapped in horovod's distributed optimizer:
                          'hvd.DistributedOptimizer'.

        :return: The updated Wrapped optimizer.

        :raise MLRunInvalidArgumentError: In case the optimizer was passed as a string.
        """
        # Check if needed to run with horovod:
        if self._hvd is None:
            return optimizer

        # Validate the optimizer input:
        if isinstance(optimizer, str):
            raise mlrun.errors.MLRunInvalidArgumentError(
                "When using horovod, the compile method is expecting an initialized optimizer instance and not a "
                "string."
            )

        # Setup the device to run on GPU if available:
        if (
            tf.config.experimental.list_physical_devices("GPU")
            and os.environ.get("CUDA_VISIBLE_DEVICES", "None") != "-1"
        ):
            # Pin each GPU to a single process:
            gpus = tf.config.experimental.list_physical_devices("GPU")
            if gpus:
                tf.config.experimental.set_visible_devices(
                    gpus[self._hvd.local_rank()], "GPU"
                )
                print(
                    f"Horovod worker #{self._hvd.rank()} is using GPU #{self._hvd.local_rank()}"
                )
        else:
            # No GPUs were found, or 'use_cuda' was false:
            os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
            print(f"Horovod worker #{self._hvd.rank()} is using CPU")

        # Adjust learning rate based on the number of GPUs:
        if hasattr(optimizer, "lr"):
            optimizer.lr = optimizer.lr * self._hvd.size()
        else:
            optimizer.learning_rate = optimizer.learning_rate * self._hvd.size()

        # Wrap the optimizer in horovod's distributed optimizer: 'hvd.DistributedOptimizer'.
        optimizer = self._hvd.DistributedOptimizer(optimizer)

        return optimizer

    def _pre_fit(
        self,
        callbacks: list[Callback],
        verbose: int,
        steps_per_epoch: Union[int, None],
        validation_steps: Union[int, None],
    ) -> tuple[list[Callback], int, Union[int, None], Union[int, None]]:
        """
        Method to call before calling 'fit' to setup the run and inputs for using horovod.

        :param callbacks:        Callbacks to use in the run. The callbacks will be split among the ranks so only
                                 certain callbacks (mainly logging and checkpoints) will be in rank 0.
        :param verbose:          Whether or not to print the progress of training. If '1' or '2' only rank 0 will be
                                 applied with the verbose.
        :param steps_per_epoch:  Amount of training steps to run in each epoch. The steps will be divided by the size of
                                 ranks (horovod workers).
        :param validation_steps: Amount of validation steps to run in each epoch. The steps will be divided by the size
                                 of ranks (horovod workers).

        :return: The updated parameters according to the used rank:
                 [0] = Callbacks list.
                 [1] = Verbose
                 [2] = Steps per epoch or None if not given.
                 [3] = Validation steps or None if not given.

        :raise MLRunInvalidArgumentError: If horovod is being used but the 'steps_per_epoch' parameter were not given.
        """
        # Check if needed to run with horovod:
        if self._hvd is None:
            return callbacks, verbose, steps_per_epoch, validation_steps

        # Validate steps provided for horovod:
        if steps_per_epoch is None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "When using Horovod, the parameter 'steps_per_epoch' must be provided to the 'fit' method in order to "
                "split the steps between the workers."
            )

        # Setup the callbacks:
        metric_average_callback = self._hvd.callbacks.MetricAverageCallback()
        metric_average_callback._supports_tf_logs = True
        horovod_callbacks = [
            self._hvd.callbacks.BroadcastGlobalVariablesCallback(0),
            metric_average_callback,
            self._hvd.callbacks.LearningRateWarmupCallback(
                initial_lr=float(
                    self.optimizer.lr
                    if hasattr(self.optimizer, "lr")
                    else self.optimizer.learning_rate
                )
            ),
        ]
        if self._hvd.rank() != 0:
            callbacks = [
                callback
                for callback in callbacks
                if type(callback).__name__ not in self._RANK_0_ONLY_CALLBACKS
            ]
        callbacks = horovod_callbacks + callbacks

        # Pick the verbose:
        if self._hvd.rank() != 0:
            verbose = 0

        # Adjust the number of steps per epoch based on the number of workers:
        steps_per_epoch = steps_per_epoch // self._hvd.size()
        if validation_steps is not None:
            validation_steps = validation_steps // self._hvd.size()

        return callbacks, verbose, steps_per_epoch, validation_steps

    def _pre_evaluate(
        self,
        callbacks: list[Callback],
        steps: Union[int, None],
    ) -> tuple[list[Callback], Union[int, None]]:
        """
        Method to call before calling 'evaluate' to setup the run and inputs for using horovod.

        :param callbacks: Callbacks to use in the run. The callbacks will be split among the ranks so only certain
                          callbacks (mainly logging and checkpoints) will be in rank 0.
        :param steps:     Amount of evaluation steps to run in each epoch. The steps will be divided by the size of
                          ranks (horovod workers).

        :return: The updated parameters according to the used rank:
                 [0] = Callbacks list.
                 [1] = Steps.

        :raise MLRunInvalidArgumentError: If horovod is being used but the 'steps' parameter were not given.
        """
        # Remove the 'auto_log' callback 'TensorboardLoggingCallback' (only relevant for training):
        callbacks = [
            callback
            for callback in callbacks
            if type(callback).__name__ != "TensorboardLoggingCallback"
        ]

        # Check if needed to run with horovod:
        if self._hvd is None:
            return callbacks, steps

        # Validate steps provided for horovod:
        if steps is None:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "When using Horovod, the parameter 'steps' must be provided to the 'evaluate' method in order to "
                "split the steps between the workers."
            )

        # Setup the callbacks:
        metric_average_callback = self._hvd.callbacks.MetricAverageCallback()
        metric_average_callback._supports_tf_logs = True
        horovod_callbacks = [
            self._hvd.callbacks.BroadcastGlobalVariablesCallback(0),
            metric_average_callback,
        ]
        if self._hvd.rank() != 0:
            callbacks = [
                callback
                for callback in callbacks
                if type(callback).__name__ not in self._RANK_0_ONLY_CALLBACKS
            ]
        callbacks = horovod_callbacks + callbacks

        # Adjust the number of steps per epoch based on the number of workers:
        steps = steps // self._hvd.size()

        return callbacks, steps



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import numpy as np
import packaging.version
import pandas as pd
import pyarrow
from pandas.io.json._table_schema import convert_pandas_type_to_json_field

import mlrun.features
from mlrun.model import ObjectList
from mlrun.utils import logger

from .data_types import InferOptions, pa_type_to_value_type, pd_schema_to_value_type

default_num_bins = 20


def infer_schema_from_df(
    df: pd.DataFrame,
    features: ObjectList,
    entities,
    timestamp_key: Optional[str] = None,
    entity_columns=None,
    options: InferOptions = InferOptions.Null,
    push_at_start: Optional[bool] = False,
):
    """infer feature set schema from dataframe"""
    timestamp_fields = []
    current_entities = list(entities.keys())
    entity_columns = entity_columns or []
    index_columns = dict()
    temp_features = ObjectList(mlrun.features.Feature)

    def upsert_entity(name, value_type):
        if name in current_entities:
            old_type = entities[name].value_type
            if old_type != value_type:
                logger.warning(
                    f"Overriding type of entity '{name}' from '{old_type}' to '{value_type}'. "
                    "This may result in errors or unusable data."
                )
                entities[name].value_type = value_type
        else:
            entities[name] = {"name": name, "value_type": value_type}

    # remove index column if no name provided
    if not df.index.name and pd.api.types.is_numeric_dtype(df.index):
        df = df.reset_index().drop("index", axis=1)

    schema = pyarrow.Schema.from_pandas(df)
    for i in range(len(schema)):
        column = schema.names[i]
        value_type = pa_type_to_value_type(schema.types[i])
        if column in df.index.names:
            index_columns[column] = value_type
            continue
        is_entity = column in entity_columns or column in current_entities
        if is_entity:
            upsert_entity(column, value_type)
        elif (
            InferOptions.get_common_options(options, InferOptions.Features)
            and column != timestamp_key
        ):
            if column in features.keys():
                features[column].value_type = value_type
            else:
                temp_features[column] = {"name": column, "value_type": value_type}
        if value_type == "datetime" and not is_entity:
            timestamp_fields.append(column)

    features.update_list(
        object_list=temp_features, push_at_start=push_at_start
    )  # Push to start of the Object list

    index_type = None
    if InferOptions.get_common_options(options, InferOptions.Index):
        # infer types of index fields
        if df.index.name:
            if df.index.name in index_columns:
                index_type = index_columns[df.index.name]
            if not index_type:
                field = convert_pandas_type_to_json_field(df.index)
                index_type = pd_schema_to_value_type(field["type"])
            # Workaround to infer a boolean index correctly, and not as 'str'.
            upsert_entity(df.index.name, index_type)
        elif df.index.nlevels > 1:
            for level, name in zip(df.index.levels, df.index.names):
                if name in index_columns:
                    index_type = index_columns[name]
                else:
                    field = convert_pandas_type_to_json_field(df.index)
                    index_type = pd_schema_to_value_type(field["type"])
                upsert_entity(name, index_type)
                if index_type == "datetime":
                    timestamp_fields.append(name)

    return timestamp_key


def get_df_stats(df, options, num_bins=None, sample_size=None):
    """get per column data stats from dataframe"""

    results_dict = {}
    if df.empty:
        return results_dict
    if sample_size and df.shape[0] > sample_size:
        df = df.sample(sample_size)

    num_bins = num_bins or default_num_bins
    if InferOptions.get_common_options(options, InferOptions.Index) and df.index.names:
        df = df.reset_index()
    # pandas 2 removes datetime_is_numeric
    # See https://github.com/mlflow/mlflow/pull/7898 for more information
    kwargs = (
        {}
        if packaging.version.Version(pd.__version__)
        >= packaging.version.Version("2.0.0rc0")
        else {"datetime_is_numeric": True}
    )
    for col, values in df.describe(include="all", **kwargs).items():
        stats_dict = {}
        for stat, val in values.dropna().items():
            if isinstance(val, (float, np.floating, np.float64)):
                stats_dict[stat] = float(val)
            elif isinstance(val, (int, np.integer, np.int64)):
                # boolean values are considered subclass of int
                if isinstance(val, bool):
                    stats_dict[stat] = bool(val)
                else:
                    stats_dict[stat] = int(val)
            else:
                stats_dict[stat] = str(val)

        if InferOptions.get_common_options(
            options, InferOptions.Histogram
        ) and pd.api.types.is_numeric_dtype(df[col]):
            # store histogram
            try:
                hist, bins = np.histogram(df[col], bins=num_bins)
                stats_dict["hist"] = [hist.tolist(), bins.tolist()]
            except Exception:
                pass

        results_dict[col] = stats_dict
    return results_dict


def get_df_preview(df, preview_lines=20):
    """capture preview data from df"""
    # record sample rows from the dataframe
    length = df.shape[0]
    shortdf = df
    if length > preview_lines:
        shortdf = df.sample(preview_lines)
    shortdf = shortdf.reset_index(inplace=False)
    return [shortdf.columns.values.tolist()] + shortdf.values.tolist()


class DFDataInfer:
    infer_schema = infer_schema_from_df
    get_preview = get_df_preview
    get_stats = get_df_stats



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import mlrun

from ..._ml_common.loggers import MLRunLogger
from .callback import CallbackEnv
from .logging_callback import LoggingCallback


class MLRunLoggingCallback(LoggingCallback):
    """
    A logging callback to collect training data into MLRun. The logging includes:

    * Per iteration chart artifacts for the metrics results.
    * Per iteration chart artifacts for the dynamic hyperparameters values.
    * Results table of the training including the static hyperparameters, and the last iteration dynamic hyperparameters
      values and metrics results.
    """

    def __init__(
        self,
        context: mlrun.MLClientCtx,
        dynamic_hyperparameters: Optional[list[str]] = None,
        static_hyperparameters: Optional[list[str]] = None,
        logging_frequency: int = 100,
    ):
        """
        Initialize an MLRun logging callback with the given configuration. All the metrics data will be collected but
        the hyperparameters to log must be given. The hyperparameters will be taken from the `params` of the model in
        each iteration.

        :param context:                 MLRun context to log to. The context parameters can be logged as static
                                        hyperparameters.
        :param dynamic_hyperparameters: If needed to track a hyperparameter dynamically (sample it each iteration) it
                                        should be passed here. The parameter expects a list of all the hyperparameters
                                        names to track our of the `params` dictionary.
        :param static_hyperparameters:  If needed to track a hyperparameter one time per run it should be passed here.
                                        The parameter expects a list of all the hyperparameters names to track our of
                                        the `params` dictionary.
        :param logging_frequency:       Per how many iterations to write the logs to MLRun (create the plots and log
                                        them and the results to MLRun). Two low frequency may slow the training time.
                                        Default: 100.
        """
        super().__init__(
            dynamic_hyperparameters=dynamic_hyperparameters,
            static_hyperparameters=static_hyperparameters,
        )

        # Replace the logger with an MLRun logger:
        del self._logger
        self._logger = MLRunLogger(context=context)

        # Store the logging frequency, it will be compared with the iteration received in the `CallbackEnv` tuple.
        self._logging_frequency = logging_frequency

    def __call__(self, env: CallbackEnv):
        """
        Log the iteration that ended and all the results it calculated.

        :param env: A named tuple passed ad the end of each iteration containing the metrics results. For more
                    information check the `Callback` doc string.
        """
        # Log the results and parameters:
        super().__call__(env=env)

        # Produce the artifacts (post iteration stage):
        if env.iteration % self._logging_frequency == 0:
            self._logger.log_iteration_to_context()

    def on_train_begin(self):
        """
        Log the context parameters when training begins.
        """
        self._logger.log_context_parameters()

    def on_train_end(self):
        """
        Log the last iteration training data into MLRun.
        """
        self._logger.log_iteration_to_context()



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from mlrun.common.runtimes.constants import MPIJobCRDVersions, MPIJobV1CleanPodPolicies
from mlrun.runtimes.mpijob.abstract import AbstractMPIJobRuntime, MPIResourceSpec


class MPIV1ResourceSpec(MPIResourceSpec):
    _dict_fields = MPIResourceSpec._dict_fields + ["clean_pod_policy"]

    def __init__(
        self,
        command=None,
        args=None,
        image=None,
        mode=None,
        volumes=None,
        volume_mounts=None,
        env=None,
        resources=None,
        default_handler=None,
        entry_points=None,
        description=None,
        workdir=None,
        replicas=None,
        image_pull_policy=None,
        service_account=None,
        build=None,
        image_pull_secret=None,
        mpi_args=None,
        clean_pod_policy=None,
        node_name=None,
        node_selector=None,
        affinity=None,
        priority_class_name=None,
        disable_auto_mount=False,
        pythonpath=None,
        tolerations=None,
        preemption_mode=None,
        security_context=None,
        state_thresholds=None,
        serving_spec=None,
        graph=None,
        parameters=None,
        track_models=None,
    ):
        super().__init__(
            command=command,
            image=image,
            mode=mode,
            build=build,
            entry_points=entry_points,
            description=description,
            workdir=workdir,
            default_handler=default_handler,
            volumes=volumes,
            volume_mounts=volume_mounts,
            env=env,
            resources=resources,
            replicas=replicas,
            image_pull_policy=image_pull_policy,
            service_account=service_account,
            image_pull_secret=image_pull_secret,
            args=args,
            mpi_args=mpi_args,
            node_name=node_name,
            node_selector=node_selector,
            affinity=affinity,
            priority_class_name=priority_class_name,
            disable_auto_mount=disable_auto_mount,
            pythonpath=pythonpath,
            tolerations=tolerations,
            preemption_mode=preemption_mode,
            security_context=security_context,
            state_thresholds=state_thresholds,
            serving_spec=serving_spec,
            graph=graph,
            parameters=parameters,
            track_models=track_models,
        )
        self.clean_pod_policy = clean_pod_policy or MPIJobV1CleanPodPolicies.default()


class MpiRuntimeV1(AbstractMPIJobRuntime):
    crd_group = "kubeflow.org"
    crd_version = MPIJobCRDVersions.v1
    crd_plural = "mpijobs"

    @property
    def spec(self) -> MPIV1ResourceSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", MPIV1ResourceSpec)



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional, Union

import mlrun
from mlrun.artifacts import Artifact

from ..._common import LoggingMode
from ..._dl_common.loggers import MLRunLogger
from ..model_handler import TFKerasModelHandler
from ..utils import TFKerasTypes
from .logging_callback import LoggingCallback


class MLRunLoggingCallback(LoggingCallback):
    """
    Callback for logging data during training / validation via mlrun's context. Each tracked hyperparameter and metrics
    results will be logged per epoch and at the end of the run the model will be saved and logged as well. Some plots
    will be available as well. To summerize, the available data in mlrun will be:

    * For each epoch:

      * Tracking table: epoch, static hyperparameters, dynamic hyperparameters, training metrics, validation metrics.
      * Per iteration (batch) chart artifacts for the training and validation metrics.

    * At the end of the run:

      * Per epoch chart artifacts for the validation summaries and dynamic hyperparameters.
      * Model is logged with all of the files and artifacts.

    All the collected data will be available in this callback post the training / validation process and can be accessed
    via the 'training_results', 'validation_results', 'static_hyperparameters', 'dynamic_hyperparameters' and
    'summaries' properties.
    """

    def __init__(
        self,
        context: mlrun.MLClientCtx,
        model_handler: TFKerasModelHandler,
        log_model_tag: str = "",
        log_model_labels: Optional[dict[str, TFKerasTypes.TrackableType]] = None,
        log_model_parameters: Optional[dict[str, TFKerasTypes.TrackableType]] = None,
        log_model_extra_data: Optional[
            dict[str, Union[TFKerasTypes.TrackableType, Artifact]]
        ] = None,
        dynamic_hyperparameters: Optional[
            dict[
                str,
                Union[list[Union[str, int]], Callable[[], TFKerasTypes.TrackableType]],
            ]
        ] = None,
        static_hyperparameters: Optional[
            dict[str, Union[TFKerasTypes, list[Union[str, int]]]]
        ] = None,
        auto_log: bool = False,
    ):
        """
        Initialize an mlrun logging callback with the given hyperparameters and logging configurations.

        :param context:                  MLRun context to log to. Its parameters will be logged automatically  if
                                         'auto_log' is True.
        :param model_handler:            A TFKerasModelHandler initialized with the model to be trained. The model must
                                         be loaded. The logs will be applied to it.
        :param log_model_tag:            Version tag to give the logged model.
        :param log_model_labels:         Labels to log with the model.
        :param log_model_parameters:     Parameters to log with the model.
        :param log_model_extra_data:     Extra data to log with the model.
        :param dynamic_hyperparameters:  If needed to track a hyperparameter dynamically (sample it each epoch) it
                                         should be passed here. The parameter expects a dictionary where the keys are
                                         the hyperparameter chosen names and the values are a key chain from the model.
                                         A key chain is a list of keys and indices to know how to access the needed
                                         hyperparameter from the model. If the hyperparameter is not of accessible from
                                         the model, a custom callable method can be passed. For example, to track the
                                         'lr' attribute of an optimizer and a custom parameter, one should pass:
                                         {
                                             "learning rate": ["optimizer", "lr"],
                                             "custom_parameter": get_custom_parameter
                                         }
        :param static_hyperparameters:   If needed to track a hyperparameter one time per run it should be passed here.
                                         The parameter expects a dictionary where the keys are the
                                         hyperparameter chosen names and the values are the hyperparameter static value
                                         or a key chain just like the dynamic hyperparameter. For example, to track the
                                         'epochs' of an experiment run, one should pass:
                                         {
                                             "epochs": 7
                                         }
        :param auto_log:                 Whether or not to enable auto logging for logging the context parameters and
                                         trying to track common static and dynamic hyperparameters such as learning
                                         rate.
        """
        super().__init__(
            dynamic_hyperparameters=dynamic_hyperparameters,
            static_hyperparameters=static_hyperparameters,
            auto_log=auto_log,
        )

        # Replace the logger with an MLRunLogger:
        del self._logger
        self._logger = MLRunLogger(context=context)

        # Store the attributes to log along the model:
        self._log_model_tag = log_model_tag
        self._log_model_labels = log_model_labels
        self._log_model_parameters = log_model_parameters
        self._log_model_extra_data = log_model_extra_data

        # Store the model handler:
        self._model_handler = model_handler

    def on_train_end(self, logs: Optional[dict] = None):
        """
        Called at the end of training, logging the model and the summaries of this run.

        :param logs: Currently the output of the last call to `on_epoch_end()` is passed to this argument for this
                     method but that may change in the future.
        """
        self._end_run()

    def on_test_end(self, logs: Optional[dict] = None):
        """
        Called at the end of evaluation or validation. Will be called on each epoch according to the validation
        per epoch configuration. The recent evaluation / validation results will be summarized and logged. If the logger
        is in evaluation mode, the model artifact will be updated.

        :param logs: Currently no data is passed to this argument for this method but that may change in the
                     future.
        """
        super().on_test_end(logs=logs)

        # Check if its part of evaluation. If so, end the run:
        if self._logger.mode == LoggingMode.EVALUATION:
            self._logger.log_epoch_to_context(epoch=1)
            self._end_run()

    def on_epoch_end(self, epoch: int, logs: Optional[dict] = None):
        """
        Called at the end of an epoch, logging the dynamic hyperparameters and results of this epoch via the stored
        context.

        :param epoch: Integer, index of epoch.
        :param logs:  Dict, metric results for this training epoch, and for the validation epoch if validation is
                      performed. Validation result keys are prefixed with `val_`. For training epoch, the values of the
                      `Model`'s metrics are returned. Example : `{'loss': 0.2, 'acc': 0.7}`.
        """
        super().on_epoch_end(epoch=epoch)

        # Log the current epoch's results:
        self._logger.log_epoch_to_context(epoch=epoch)

    def _end_run(self):
        """
        End the run, logging the collected artifacts.
        """
        # Set the inputs information if needed:
        if self._model_handler.inputs is None:
            self._model_handler.read_inputs_from_model()

        # Set the outputs information if needed:
        if self._model_handler.outputs is None:
            self._model_handler.read_outputs_from_model()

        # Log the model:
        self._logger.log_run(
            model_handler=self._model_handler,
            tag=self._log_model_tag,
            labels=self._log_model_labels,
            parameters=self._log_model_parameters,
            extra_data=self._log_model_extra_data,
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "BaseRuntime",
    "KubejobRuntime",
    "LocalRuntime",
    "HandlerRuntime",
    "RemoteRuntime",
    "ServingRuntime",
    "DaskCluster",
    "RemoteSparkRuntime",
    "Spark3Runtime",
    "DatabricksRuntime",
    "KubeResource",
    "ApplicationRuntime",
    "MpiRuntimeV1",
]

import typing

from mlrun.runtimes.utils import resolve_spark_operator_version

from ..common.runtimes.constants import MPIJobCRDVersions
from .base import BaseRuntime, RunError, RuntimeClassMode  # noqa
from .daskjob import DaskCluster  # noqa
from .databricks_job.databricks_runtime import DatabricksRuntime
from .kubejob import KubejobRuntime, KubeResource  # noqa
from .local import HandlerRuntime, LocalRuntime  # noqa
from .mpijob import MpiRuntimeV1  # noqa
from .nuclio import (
    RemoteRuntime,
    ServingRuntime,
    new_v2_model_server,
    nuclio_init_hook,
)
from .nuclio.api_gateway import APIGateway
from .nuclio.application import ApplicationRuntime
from .nuclio.serving import serving_subkind
from .remotesparkjob import RemoteSparkRuntime
from .sparkjob import Spark3Runtime

# for legacy imports (MLModelServer moved from here to /serving)
from ..serving import MLModelServer, new_v1_model_server  # noqa isort: skip


def new_model_server(
    name,
    model_class: str,
    models: typing.Optional[dict] = None,
    filename="",
    protocol="",
    image="",
    endpoint="",
    explainer=False,
    workers=8,
    canary=None,
    handler=None,
):
    if protocol:
        return new_v2_model_server(
            name,
            model_class,
            models=models,
            filename=filename,
            protocol=protocol,
            image=image,
            endpoint=endpoint,
            workers=workers,
            canary=canary,
        )
    else:
        return new_v1_model_server(
            name,
            model_class,
            models=models,
            filename=filename,
            protocol=protocol,
            image=image,
            endpoint=endpoint,
            workers=workers,
            canary=canary,
        )


class RuntimeKinds:
    remote = "remote"
    nuclio = "nuclio"
    dask = "dask"
    job = "job"
    spark = "spark"
    remotespark = "remote-spark"
    mpijob = "mpijob"
    serving = "serving"
    local = "local"
    handler = "handler"
    databricks = "databricks"
    application = "application"

    @staticmethod
    def all():
        return [
            RuntimeKinds.remote,
            RuntimeKinds.nuclio,
            RuntimeKinds.serving,
            RuntimeKinds.dask,
            RuntimeKinds.job,
            RuntimeKinds.spark,
            RuntimeKinds.remotespark,
            RuntimeKinds.mpijob,
            RuntimeKinds.local,
            RuntimeKinds.databricks,
            RuntimeKinds.application,
        ]

    @staticmethod
    def runtime_with_handlers():
        return [
            RuntimeKinds.dask,
            RuntimeKinds.job,
            RuntimeKinds.spark,
            RuntimeKinds.remotespark,
            RuntimeKinds.mpijob,
            RuntimeKinds.databricks,
        ]

    @staticmethod
    def abortable_runtimes():
        return [
            RuntimeKinds.job,
            RuntimeKinds.spark,
            RuntimeKinds.remotespark,
            RuntimeKinds.mpijob,
            RuntimeKinds.databricks,
            RuntimeKinds.local,
            RuntimeKinds.handler,
            "",
        ]

    @staticmethod
    def retriable_runtimes():
        return [
            RuntimeKinds.job,
        ]

    @staticmethod
    def nuclio_runtimes():
        return [
            RuntimeKinds.remote,
            RuntimeKinds.nuclio,
            RuntimeKinds.serving,
            RuntimeKinds.application,
        ]

    @staticmethod
    def pure_nuclio_deployed_runtimes():
        return [
            RuntimeKinds.remote,
            RuntimeKinds.nuclio,
            RuntimeKinds.serving,
        ]

    @staticmethod
    def handlerless_runtimes():
        return [
            RuntimeKinds.serving,
            # Application runtime handler is internal reverse proxy
            RuntimeKinds.application,
        ]

    @staticmethod
    def local_runtimes():
        return [
            RuntimeKinds.local,
            RuntimeKinds.handler,
        ]

    @staticmethod
    def is_log_collectable_runtime(kind: typing.Optional[str]):
        """
        whether log collector can collect logs for that runtime
        :param kind: kind name
        :return: whether log collector can collect logs for that runtime
        """
        # if local run, the log collector doesn't support it as it is only supports k8s resources
        # when runtime is local the client is responsible for logging the stdout of the run by using `log_std`
        if RuntimeKinds.is_local_runtime(kind):
            return False

        if (
            kind
            not in [
                # dask implementation is different from other runtimes, because few runs can be run against the same
                # runtime resource, so collecting logs on that runtime resource won't be correct, the way we collect
                # logs for dask is by using `log_std` on client side after we execute the code against the cluster,
                # as submitting the run with the dask client will return the run stdout.
                # For more information head to `DaskCluster._run`.
                RuntimeKinds.dask
            ]
            + RuntimeKinds.nuclio_runtimes()
        ):
            return True

        return False

    @staticmethod
    def is_local_runtime(kind):
        # "" or None counted as local
        if not kind or kind in RuntimeKinds.local_runtimes():
            return True
        return False

    @staticmethod
    def requires_k8s_name_validation(kind: str) -> bool:
        """
        Returns True if the runtime kind creates Kubernetes resources that use the function name.

        Function names for k8s-deployed runtimes must conform to DNS-1123 label requirements:
        - Lowercase alphanumeric characters or '-'
        - Start and end with an alphanumeric character
        - Maximum 63 characters

        Local runtimes (local, handler) run on the local machine and don't create k8s resources,
        so they don't require k8s naming validation.

        :param kind: Runtime kind string (job, spark, serving, local, etc.)
        :return: True if function name needs k8s DNS-1123 validation, False otherwise
        """
        return not RuntimeKinds.is_local_runtime(kind)

    @staticmethod
    def requires_absolute_artifacts_path(kind):
        """
        Returns True if the runtime kind requires absolute artifacts' path (i.e. is local), False otherwise.
        """
        if RuntimeKinds.is_local_runtime(kind):
            return False

        if kind not in [
            # logging artifacts is done externally to the dask cluster by a client that can either run locally (in which
            # case the path can be relative) or remotely (in which case the path must be absolute and will be passed
            # to another run)
            RuntimeKinds.dask
        ]:
            return True
        return False

    @staticmethod
    def requires_image_name_for_execution(kind):
        if RuntimeKinds.is_local_runtime(kind):
            return False

        # both spark and remote spark uses different mechanism for assigning images
        return kind not in [RuntimeKinds.spark, RuntimeKinds.remotespark]

    @staticmethod
    def supports_from_notebook(kind):
        return kind not in [RuntimeKinds.application]

    @staticmethod
    def resolve_nuclio_runtime(kind: str, sub_kind: str):
        kind = kind.split(":")[0]
        if kind not in RuntimeKinds.nuclio_runtimes():
            raise ValueError(
                f"Kind {kind} is not a nuclio runtime, available runtimes are {RuntimeKinds.nuclio_runtimes()}"
            )

        if sub_kind == serving_subkind:
            return ServingRuntime()

        if kind == RuntimeKinds.application:
            return ApplicationRuntime()

        runtime = RemoteRuntime()
        runtime.spec.function_kind = sub_kind
        return runtime

    @staticmethod
    def resolve_nuclio_sub_kind(kind):
        is_nuclio = kind.startswith("nuclio")
        sub_kind = kind[kind.find(":") + 1 :] if is_nuclio and ":" in kind else None
        if kind == RuntimeKinds.serving:
            is_nuclio = True
            sub_kind = serving_subkind
        elif kind == RuntimeKinds.application:
            is_nuclio = True
        return is_nuclio, sub_kind


def get_runtime_class(kind: str):
    if kind == RuntimeKinds.mpijob:
        return MpiRuntimeV1

    if kind == RuntimeKinds.spark:
        return Spark3Runtime

    kind_runtime_map = {
        RuntimeKinds.remote: RemoteRuntime,
        RuntimeKinds.nuclio: RemoteRuntime,
        RuntimeKinds.serving: ServingRuntime,
        RuntimeKinds.dask: DaskCluster,
        RuntimeKinds.job: KubejobRuntime,
        RuntimeKinds.local: LocalRuntime,
        RuntimeKinds.remotespark: RemoteSparkRuntime,
        RuntimeKinds.databricks: DatabricksRuntime,
        RuntimeKinds.application: ApplicationRuntime,
    }

    return kind_runtime_map[kind]



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum
import getpass
import hashlib
import json
import os
import re
from io import StringIO
from sys import stderr
from typing import Optional

import pandas as pd

import mlrun
import mlrun.common.constants
import mlrun.common.constants as mlrun_constants
import mlrun.common.runtimes.constants
import mlrun.common.schemas
import mlrun.utils.regex
from mlrun.artifacts import TableArtifact
from mlrun.config import config
from mlrun.errors import err_to_str
from mlrun.frameworks.parallel_coordinates import gen_pcp_plot
from mlrun.runtimes.generators import selector
from mlrun.utils import get_in, helpers, logger, verify_field_regex


class RunError(Exception):
    pass


class _ContextStore:
    def __init__(self):
        self._context = None

    def get(self):
        return self._context

    def set(self, context):
        self._context = context


global_context = _ContextStore()


def resolve_spark_operator_version():
    try:
        regex = re.compile("spark-([23])")
        return int(regex.findall(config.spark_operator_version)[0])
    except Exception:
        raise ValueError("Failed to resolve spark operator's version")


def calc_hash(func, tag=""):
    # remove tag, hash, date from calculation
    tag = tag or func.metadata.tag
    status = func.status
    func.metadata.tag = ""
    func.metadata.hash = ""
    func.status = None
    func.metadata.updated = None

    data = json.dumps(func.to_dict(), sort_keys=True).encode()
    h = hashlib.sha1()
    h.update(data)
    hashkey = h.hexdigest()
    func.metadata.tag = tag
    func.metadata.hash = hashkey
    func.status = status
    return hashkey


def log_std(db, runobj, out, err="", skip=False, show=True, silent=False):
    if out:
        iteration = runobj.metadata.iteration
        if iteration:
            line = "> " + "-" * 15 + f" Iteration: ({iteration}) " + "-" * 15 + "\n"
            out = line + out
        if show:
            print(out, flush=True)
        if db and not skip:
            uid = runobj.metadata.uid
            project = runobj.metadata.project or ""
            db.store_log(uid, project, out.encode(), append=True)
    if err:
        logger.error(f"Exec error - {err_to_str(err)}")
        print(err, file=stderr)
        if not silent:
            raise RunError(err)


def add_code_metadata(path=""):
    if path:
        if "://" in path:
            return None
        if os.path.isfile(path):
            path = os.path.dirname(path)
    path = path or "./"

    try:
        from git import (
            GitCommandNotFound,
            InvalidGitRepositoryError,
            NoSuchPathError,
            Repo,
        )
    except ImportError:
        return None

    try:
        repo = Repo(path, search_parent_directories=True)
        remotes = [
            remote.url
            for remote in repo.remotes
            # some stale remotes might be missing urls
            # there is not a nicer way to check for this.
            if repo.config_reader().has_option(f'remote "{remote}"', "url")
        ]
        if len(remotes) > 0:
            return f"{remotes[0]}#{repo.head.commit.hexsha}"

    except (
        InvalidGitRepositoryError,
        NoSuchPathError,
    ):
        # Path is not part of a git repository or an invalid path (will fail later if it needs to)
        pass

    except (GitCommandNotFound, ValueError) as exc:
        logger.warning(
            "Failed to add git metadata",
            path=path,
            error=err_to_str(exc),
        )
    return None


def results_to_iter(results, runspec, execution):
    if not results:
        logger.error("Got an empty results list in to_iter")
        return

    iter = []
    failed = 0
    pending_retry = 0
    running = 0
    for task in results:
        if task:
            state = get_in(task, ["status", "state"])
            id = get_in(task, ["metadata", "iteration"])
            struct = {
                "param": get_in(task, ["spec", "parameters"], {}),
                "output": get_in(task, ["status", "results"], {}),
                "state": state,
                "iter": id,
            }
            if state == mlrun.common.runtimes.constants.RunStates.error:
                failed += 1
                err = get_in(task, ["status", "error"], "")
                logger.error(f"error in task {execution.uid}:{id} - {err_to_str(err)}")
            elif state == mlrun.common.runtimes.constants.RunStates.pending_retry:
                pending_retry += 1
                err = get_in(task, ["status", "error"], "")
                retry_count = get_in(task, ["status", "retry_count"], 0)
                logger.warning(
                    f"pending retry in task {execution.uid}:{id} - {err_to_str(err)}. Retry count: {retry_count}"
                )
            elif state != mlrun.common.runtimes.constants.RunStates.completed:
                running += 1

            iter.append(struct)

    if not iter:
        execution.set_state(
            mlrun.common.runtimes.constants.RunStates.completed, commit=True
        )
        logger.warning("Warning!, zero iteration results")
        return
    if hasattr(pd, "json_normalize"):
        df = pd.json_normalize(iter).sort_values("iter")
    else:
        df = pd.io.json.json_normalize(iter).sort_values("iter")
    header = df.columns.values.tolist()
    summary = [header] + df.values.tolist()
    if not runspec:
        return summary, df

    criteria = runspec.spec.hyper_param_options.selector
    item, id = selector(results, criteria)
    if runspec.spec.selector and not id:
        logger.warning(
            f"No best result selected, check selector ({criteria}) or results"
        )
    if id:
        logger.info(f"Best iteration={id}, used criteria {criteria}")
    task = results[item] if id and results else None
    execution.log_iteration_results(id, summary, task)

    log_iter_artifacts(execution, df, header)

    if failed:
        execution.set_state(
            error=f"{failed} of {len(results)} tasks failed, check logs in db for details",
            commit=False,
        )
    elif pending_retry:
        execution.set_state(
            mlrun.common.runtimes.constants.RunStates.pending_retry, commit=False
        )
    elif running == 0:
        execution.set_state(
            mlrun.common.runtimes.constants.RunStates.completed, commit=False
        )
    execution.commit()


def log_iter_artifacts(execution, df, header):
    csv_buffer = StringIO()
    df.to_csv(
        csv_buffer,
        index=False,
        encoding="utf-8",
        **mlrun.utils.line_terminator_kwargs(),
    )
    try:
        # may fail due to lack of access credentials to the artifacts store
        execution.log_artifact(
            TableArtifact(
                "iteration_results",
                body=csv_buffer.getvalue(),
                header=header,
                viewer="table",
            ),
            local_path="iteration_results.csv",
        )
        # may also fail due to missing plotly
        execution.log_artifact(
            "parallel_coordinates",
            body=gen_pcp_plot(df, index_col="iter"),
            local_path="parallel_coordinates.html",
        )
    except Exception as exc:
        logger.warning(f"failed to log iter artifacts, {err_to_str(exc)}")


def fill_function_image_name_template(
    registry: str,
    repository: str,
    project: str,
    name: str,
    tag: str,
) -> str:
    image_name_prefix = resolve_function_target_image_name_prefix(project, name)
    return f"{registry}{repository}/{image_name_prefix}:{tag}"


def resolve_function_target_image_name_prefix(project: str, name: str):
    return config.httpdb.builder.function_target_image_name_prefix_template.format(
        project=project, name=name
    )


def resolve_function_target_image_registries_to_enforce_prefix():
    registry, repository = helpers.get_parsed_docker_registry()
    repository = helpers.get_docker_repository_or_default(repository)
    return [
        f"{mlrun.common.constants.IMAGE_NAME_ENRICH_REGISTRY_PREFIX}{repository}/",
        f"{registry}/{repository}/",
    ]


def set_named_item(obj, item):
    if isinstance(item, dict):
        obj[item["name"]] = item
    else:
        obj[item.name] = item


def get_item_name(item, attr="name"):
    if isinstance(item, dict):
        return item.get(attr)
    else:
        return getattr(item, attr, None)


def verify_limits(
    resources_field_name,
    mem=None,
    cpu=None,
    gpus=None,
    gpu_type="nvidia.com/gpu",
):
    if mem:
        verify_field_regex(
            f"function.spec.{resources_field_name}.limits.memory",
            mem,
            mlrun.utils.regex.k8s_resource_quantity_regex
            + mlrun.utils.regex.pipeline_param,
            mode=mlrun.common.schemas.RegexMatchModes.any,
        )
    if cpu:
        verify_field_regex(
            f"function.spec.{resources_field_name}.limits.cpu",
            cpu,
            mlrun.utils.regex.k8s_resource_quantity_regex
            + mlrun.utils.regex.pipeline_param,
            mode=mlrun.common.schemas.RegexMatchModes.any,
        )
    # https://kubernetes.io/docs/tasks/manage-gpus/scheduling-gpus/
    if gpus:
        verify_field_regex(
            f"function.spec.{resources_field_name}.limits.gpus",
            gpus,
            mlrun.utils.regex.k8s_resource_quantity_regex
            + mlrun.utils.regex.pipeline_param,
            mode=mlrun.common.schemas.RegexMatchModes.any,
        )
    return generate_resources(mem=mem, cpu=cpu, gpus=gpus, gpu_type=gpu_type)


def verify_requests(
    resources_field_name,
    mem=None,
    cpu=None,
):
    if mem:
        verify_field_regex(
            f"function.spec.{resources_field_name}.requests.memory",
            mem,
            mlrun.utils.regex.k8s_resource_quantity_regex
            + mlrun.utils.regex.pipeline_param,
            mode=mlrun.common.schemas.RegexMatchModes.any,
        )
    if cpu:
        verify_field_regex(
            f"function.spec.{resources_field_name}.requests.cpu",
            cpu,
            mlrun.utils.regex.k8s_resource_quantity_regex
            + mlrun.utils.regex.pipeline_param,
            mode=mlrun.common.schemas.RegexMatchModes.any,
        )
    return generate_resources(mem=mem, cpu=cpu)


def get_gpu_from_resource_requirement(requirement: dict):
    """
    Because there could be different types of gpu types, and we don't know all the gpu types possible,
    we want to get the gpu type and its value, we can figure out the type by knowing what resource types are static
    and the possible number of resources.
    Kubernetes support 3 types of resources, two of which their name doesn't change : cpu, memory.
    :param requirement: requirement resource ( limits / requests ) which contain the resources.
    """
    if not requirement:
        return None, None

    if len(requirement) > 3:
        raise mlrun.errors.MLRunInvalidArgumentError(
            "Unable to resolve the gpu type because there are more than 3 resources"
        )
    for resource, value in requirement.items():
        if resource not in ["cpu", "memory"]:
            return resource, value
    return None, None


def generate_resources(mem=None, cpu=None, gpus=None, gpu_type="nvidia.com/gpu"):
    """get pod cpu/memory/gpu resources dict"""
    resources = {}
    if gpus:
        resources[gpu_type] = gpus
    if mem:
        resources["memory"] = mem
    if cpu:
        resources["cpu"] = cpu
    return resources


def get_func_selector(project, name=None, tag=None):
    s = [f"{mlrun_constants.MLRunInternalLabels.project}={project}"]
    if name:
        s.append(f"{mlrun_constants.MLRunInternalLabels.function}={name}")
        s.append(f"{mlrun_constants.MLRunInternalLabels.tag}={tag or 'latest'}")
    return s


def enrich_function_from_dict(function, function_dict):
    override_function = mlrun.new_function(runtime=function_dict, kind=function.kind)
    for attribute in [
        "volumes",
        "volume_mounts",
        "env",
        "resources",
        "image",
        "image_pull_policy",
        "replicas",
        "node_name",
        "node_selector",
        "affinity",
        "priority_class_name",
        "credentials",
        "tolerations",
        "preemption_mode",
        "security_context",
    ]:
        if attribute == "credentials":
            override_value = getattr(override_function.metadata, attribute, None)
        else:
            override_value = getattr(override_function.spec, attribute, None)
        if override_value:
            if attribute == "env":
                for env_dict in override_value:
                    if env_dict.get("value") is not None:
                        function.set_env(env_dict["name"], env_dict["value"])
                    else:
                        function.set_env(
                            env_dict["name"],
                            value_from=env_dict["valueFrom"],
                        )
            elif attribute == "volumes":
                function.spec.update_vols_and_mounts(override_value, [])
            elif attribute == "volume_mounts":
                # volume mounts don't have a well defined identifier (like name for volume) so we can't merge,
                # only override
                function.spec.volume_mounts = override_value
            elif attribute == "resources":
                # don't override if there are limits and requests but both are empty
                if override_value.get("limits", {}) or override_value.get(
                    "requests", {}
                ):
                    setattr(function.spec, attribute, override_value)
            elif attribute == "credentials":
                if any(override_value.to_dict().values()):
                    function.metadata.credentials = override_value
            else:
                setattr(function.spec, attribute, override_value)
    return function


def resolve_owner(
    labels: dict,
    owner_to_enrich: Optional[str] = None,
):
    """
    Resolve the owner label value
    :param labels: The run labels dict
    :param auth_username: The authenticated username
    :return: The resolved owner label value
    """

    if owner_to_enrich and (
        labels.get("job-type") == mlrun.common.constants.JOB_TYPE_WORKFLOW_RUNNER
        or labels.get("job-type")
        == mlrun.common.constants.JOB_TYPE_RERUN_WORKFLOW_RUNNER
    ):
        return owner_to_enrich
    else:
        return os.environ.get("V3IO_USERNAME") or getpass.getuser()


def enrich_run_labels(
    labels: dict,
    labels_to_enrich: Optional[list[mlrun_constants.MLRunInternalLabels]] = None,
    owner_to_enrich: Optional[str] = None,
):
    """
    Enrich the run labels with the internal labels and the labels enrichment extension.
    :param labels: The run labels dict
    :param labels_to_enrich: The label keys to enrich from MLRunInternalLabels.default_run_labels_to_enrich
    :param owner_to_enrich: Optional owner to enrich the labels with, if not provided will try to resolve it.
    :return: The enriched labels dict
    """
    # Merge the labels with the labels enrichment extension
    labels_enrichment = {
        mlrun_constants.MLRunInternalLabels.owner: resolve_owner(
            labels, owner_to_enrich
        ),
    }
    # Resolve which label keys to enrich
    if labels_to_enrich is None:
        labels_to_enrich = (
            mlrun_constants.MLRunInternalLabels.default_run_labels_to_enrich()
        )

    # Enrich labels
    for label in labels_to_enrich:
        if isinstance(label, enum.Enum):
            label = label.value
        enrichment = labels_enrichment.get(label)
        if label not in labels and enrichment:
            labels[label] = enrichment
    return labels


def resolve_node_selectors(
    project_node_selector: dict, instance_node_selector: dict
) -> dict:
    config_node_selector = mlrun.mlconf.get_default_function_node_selector()
    if project_node_selector or config_node_selector:
        mlrun.utils.logger.debug(
            "Enriching node selector from project and mlrun config",
            project_node_selector=project_node_selector,
            config_node_selector=config_node_selector,
        )
        return mlrun.utils.helpers.merge_dicts_with_precedence(
            config_node_selector,
            project_node_selector,
            instance_node_selector,
        )
    return instance_node_selector


def enrich_gateway_timeout_annotations(annotations: dict, gateway_timeout: int):
    """
    Set gateway proxy connect/read/send timeout annotations
    :param annotations:     The annotations to enrich
    :param gateway_timeout: The timeout to set
    """
    if not gateway_timeout:
        return
    gateway_timeout_str = str(gateway_timeout)
    annotations["nginx.ingress.kubernetes.io/proxy-connect-timeout"] = (
        gateway_timeout_str
    )
    annotations["nginx.ingress.kubernetes.io/proxy-read-timeout"] = gateway_timeout_str
    annotations["nginx.ingress.kubernetes.io/proxy-send-timeout"] = gateway_timeout_str



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Callable, Optional, Union

import numpy as np


class ONNXDataset:
    """
    An iterable class handling numpy arrays for ONNX inference.
    """

    def __init__(
        self,
        x: Union[np.ndarray, list[np.ndarray]],
        y: Union[np.ndarray, list[np.ndarray]] = None,
        batch_size: int = 1,
        x_transforms: Optional[list[Callable[[np.ndarray], np.ndarray]]] = None,
        y_transforms: Optional[list[Callable[[np.ndarray], np.ndarray]]] = None,
        is_batched_transforms: bool = False,
    ):
        """
        Initialize a new ONNX dataset.

        :param x:                     The data to handle.
        :param y:                     Ground truth for the data 'x'.
        :param batch_size:            The batch size to use. Default: 1.
        :param x_transforms:          A list of callable transforms to apply on the 'x' data.
        :param y_transforms:          A list of callable transforms to apply on the 'y' ground truth.
        :param is_batched_transforms: Whether the given transforms support batched data or not. If not, the transform
                                      will be applied on each item of a given batch. Default: False.
        """
        # Store given parameters:
        self._x = x
        self._y = y
        self._batch_size = batch_size
        self._x_transforms = x_transforms
        self._y_transforms = y_transforms
        self._is_batched_transforms = is_batched_transforms

        # Set inner index to 0:
        self._index = 0

    def __len__(self):
        """
        Calculate the dataset's length in batches.

        :return: The dataset's length.
        """
        return math.ceil(len(self._x) / self._batch_size)

    def __iter__(self):
        """
        Set the inner index of the dataset to 0 and return the dataset as the iterator.

        :return: The dataset itself.
        """
        self._index = 0
        return self

    def __next__(self) -> Union[np.ndarray, tuple[np.ndarray, np.ndarray]]:
        """
        Get the next item in line (by the inner index) since calling '__iter__'. If ground truth was provided (y),
        a tuple of (x, y) will be returned. Otherwise x.

        :return: The next item.

        :raise StopIteration: If there are no more items to be returned.
        """
        # Validate the next item is in range:
        if self._index >= len(self):
            raise StopIteration

        # Get the next item:
        item = self[self._index]
        self._index += 1

        return item

    def __getitem__(
        self, index: int
    ) -> Union[np.ndarray, tuple[np.ndarray, np.ndarray]]:
        """
        Get the item at the given index. If ground truth was provided, a tuple of (x, y) will be returned. Otherwise x.

        :param index: The index to get the item. Must be in range of 0 to dataset's length.

        :return: The item at the given index.

        :raise IndexError: If the given index is not in range.
        """
        # Calculate the indexes:
        from_index = self._batch_size * index
        if from_index >= len(self):
            raise IndexError(
                f"Dataset's length is {len(self)}, yet given index was {index}."
            )
        to_index = max(self._batch_size * (index + 1), len(self._x))
        is_batched = (to_index - from_index) > 1

        # Get the index data:
        x = self._x[from_index:to_index]
        x = self._call_x_transforms(x=x, is_batched=is_batched)
        if self._y is None:
            return x

        # Get the index ground truth:
        y = self._y[from_index:to_index]
        y = self._call_y_transforms(y=y, is_batched=is_batched)
        return x, y

    def _call_x_transforms(self, x: np.ndarray, is_batched: bool):
        """
        Call the data transforms on the given input.

        :param x:          The input to get trough the data transforms.
        :param is_batched: Whether or not the given input is batched.

        :return: Transformed input.
        """
        if self._x_transforms is not None:
            return self._call_transforms(
                items=x, transforms=self._x_transforms, is_batched=is_batched
            )
        return x

    def _call_y_transforms(self, y: np.ndarray, is_batched: bool):
        """
        Call the ground truth transforms on the given input.

        :param y:          The input to get trough the ground truth transforms.
        :param is_batched: Whether or not the given input is batched.

        :return: Transformed input.
        """
        if self._y_transforms is not None:
            return self._call_transforms(
                items=y, transforms=self._y_transforms, is_batched=is_batched
            )
        return y

    def _call_transforms(
        self,
        items: np.ndarray,
        transforms: list[Callable[[np.ndarray], np.ndarray]],
        is_batched: bool,
    ):
        """
        Propagate the given items through the given transforms.

        :param items:      The items to transform.
        :param transforms: The transforms to use.
        :param is_batched: Whether or not the items are in a batch or not (single item).

        :return: The transformed items.
        """
        if self._is_batched_transforms or not is_batched:
            for transform in transforms:
                items = transform(items)
        else:
            for transform in transforms:
                items = [transform(item) for item in items]

        return items



# Copyright 2018 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import pathlib
import tempfile
import zipfile
from typing import Optional

import mlflow
import mlflow.entities
import mlflow.environment_variables
import mlflow.types

import mlrun
from mlrun import MLClientCtx, mlconf
from mlrun.artifacts import Artifact, ModelArtifact
from mlrun.features import Feature
from mlrun.launcher.client import ClientBaseLauncher
from mlrun.model import RunObject
from mlrun.projects import MlrunProject
from mlrun.track.tracker import Tracker
from mlrun.utils import logger, now_date


class MLFlowTracker(Tracker):
    """
    A class for detecting and logging MLFlow runs into MLRun. Notice, only the last active MLFlow run is logged.
    """

    def __init__(self):
        super().__init__()
        self._pre_runs = None
        self._experiment_name = None
        self._run_id = None

    @staticmethod
    def is_enabled() -> bool:
        """
        Checks if tracker is enabled.

        :return: True if the tracking configuration is enabled, False otherwise.
        """
        return mlconf.external_platform_tracking.mlflow.enabled

    def pre_run(self, context: MLClientCtx):
        """
        Initializes the MLFlow tracker, setting the experiment name as configured by the user.

        :param context: Current mlrun context
        """
        # Check for a user set experiment name via the environment variable:
        self._experiment_name = (
            mlflow.environment_variables.MLFLOW_EXPERIMENT_NAME.get()
        )

        # Check if the user configured for matching the experiment name with the context name:
        if mlconf.external_platform_tracking.mlflow.match_experiment_to_runtime:
            if self._experiment_name is not None:
                context.logger.warn(
                    f"`mlconf.external_platform_tracking.mlflow.match_experiment_to_runtime` is set to True but the "
                    f"MLFlow experiment name environment variable ('MLFLOW_EXPERIMENT_NAME') is set for using the "
                    f"name: '{self._experiment_name}'. This name will be overriden "
                    f"with MLRun's runtime name as set in the MLRun configuration: '{context.name}'."
                )
            self._experiment_name = context.name
            mlflow.environment_variables.MLFLOW_EXPERIMENT_NAME.set(
                self._experiment_name
            )

        # We have 3 options to track our run:
        # 1. Set the run id.
        # 2. Use the experiment, and check for added runs in it.
        # 3. Look for added runs across all experiments.
        # In case we want to determine the run id we have to create a new mlflow run, save the run id and use it
        # to override the run later (due to mlflow limitations)
        if mlrun.mlconf.external_platform_tracking.mlflow.control_run:
            with mlflow.start_run() as run:
                self._run_id = run.info.run_id
            mlflow.environment_variables.MLFLOW_RUN_ID.set(self._run_id)

        elif self._experiment_name:
            # We try to load an existing experiment first
            experiment = mlflow.get_experiment_by_name(name=self._experiment_name)
            # If no experiment with corresponding name exists, we create a new one
            if not experiment:
                experiment_id = mlflow.create_experiment(name=self._experiment_name)
                experiment = mlflow.get_experiment(experiment_id=experiment_id)

            # Save all runs logged in experiment prior of our run for later comparison
            self._pre_runs = set(
                [
                    run.info.run_id
                    for run in mlflow.search_runs(
                        experiment_ids=experiment.experiment_id,
                        output_format="list",
                    )
                ]
            )

        else:  # TODO Talk about removing/replacing with more exact methods
            # We warn user that when no experiment name or run id is given we look at all runs and might track the wrong
            # run in case where number of runs are running simultaneously
            context.logger.warn(
                "MLFlow Experiment Name and MLFlow Run ID Not Found. "
                "In this case, MLRun encountered an issue as it was unable to locate the MLFlow"
                " experiment name and run ID. When this occurs, MLRun will attempt to search through "
                "all experiments and runs to identify the newly added run. However, in scenarios where"
                " a substantial number of runs are active simultaneously, MLRun may not accurately log"
                " the correct run. To resolve this issue, please consider the following steps: "
                "1. Set MLFlow Environment Variables: Ensure that you have set the necessary"
                " MLFlow environment variables, specifically 'MLFLOW_EXPERIMENT_NAME' and"
                " 'MLFLOW_RUN_ID', to provide explicit identification for the experiment and run. "
                "2. Adjust mlconf Configuration (mlconf): "
                "Alternatively, you can configure MLRun by setting "
                "'mlconf.external_platform_tracking.mlflow.match_experiment_to_runtime` to True within"
                " the mlrun configuration (mlconf). This setting enables MLRun to match the "
                "experiment with the runtime more effectively. "
            )
            # Find all experiments to save all their runs for later comparison
            all_experiments = [exp.experiment_id for exp in mlflow.search_experiments()]
            self._pre_runs = set(
                [
                    run.info.run_id
                    for run in mlflow.search_runs(
                        experiment_ids=all_experiments,
                        output_format="list",
                    )
                ]
            )

    def post_run(self, context: MLClientCtx):
        """
        Performs post-run tasks for logging MLFlow artifacts generated during the run.

        :param context: Current mlrun context
        """
        # We try to find the correct run to log via available data
        if self._run_id:
            run = [self._run_id]

        elif self._experiment_name:
            # If we now the experiment's name we can look at all runs and find the new one
            experiment = mlflow.get_experiment_by_name(self._experiment_name)
            run = (
                set(
                    [
                        run.info.run_id
                        for run in mlflow.search_runs(
                            experiment_ids=experiment.experiment_id,
                            output_format="list",
                        )
                    ]
                )
                ^ self._pre_runs  # xor to find the run added to the list
            )

        else:
            # In case we don't have the experiment name or run id we need to look at all new runs to found the one added
            all_experiments = [exp.experiment_id for exp in mlflow.search_experiments()]
            run = (
                set(
                    [
                        run.info.run_id
                        for run in mlflow.search_runs(
                            experiment_ids=all_experiments,
                            output_format="list",
                        )
                    ]
                )
                ^ self._pre_runs  # xor to find the run added to the list
            )
        # should be only one run id in run
        run = mlflow.get_run(run_id=run.pop())

        # Log the run:
        MLFlowTracker._log_run(context=context, run=run)

    def import_run(
        self,
        project: MlrunProject,
        reference_id: str,
        function_name: str,
        handler: Optional[str] = None,
        **kwargs,
    ) -> RunObject:
        """
        Import a previous MLFlow experiment run to MLRun.

        :param project:       The MLRun project to import the run to.
        :param reference_id:  The MLFlow `run_id` to import.
        :param function_name: The MLRun function to assign this run to.
        :param handler:       The handler for MLRun's RunObject

        :return: The newly imported RunObject.
        """

        run = mlflow.get_run(reference_id)
        if not run:
            raise ValueError(
                "The provided run id was not found in the set MLFlow registry. Try to set the registry using "
                "MLFlow's environment variables."
            )

        # Create the run object to hold the mlflow run:
        run_object = self._create_run_object(
            handler=handler,
            run_name=run.info.run_name,
            project_name=project.name,
            uid=run.info.run_uuid,
        )

        # Create a context from the run object:
        ctx = mlrun.get_or_create_ctx(
            name=run.info.run_name,
            spec=run_object,
        )
        # Store the run in the MLRun DB, then import the MLFlow data to it:
        self._log_run(context=ctx, run=run, is_offline=True)

        # Create a rundb in order to update the run's state as completed (can't be done using context)
        rundb = mlrun.get_run_db()
        project = run_object.metadata.project
        uid = run_object.metadata.uid
        updates = {
            "status.last_update": now_date().isoformat(),
            "status.state": "completed",
        }
        run_object.status.state = "completed"
        rundb.update_run(updates, uid, project)

        # Print a summary message after importing the run:
        result = ctx.to_dict()
        run_object = RunObject.from_dict(result)
        ClientBaseLauncher._log_track_results(
            is_child=False, result=result, run=run_object
        )

        return run_object

    def import_model(
        self,
        project: MlrunProject,
        reference_id: str,
        key: Optional[str] = None,
        metrics: Optional[dict] = None,
        extra_data: Optional[dict] = None,
    ) -> ModelArtifact:
        """
        Import a model from MLFlow to MLRun.

        :param project:         The MLRun project to import the model to.
        :param reference_id:    The MLFlow model uri to import.
        :param key:             The model key to be used in MLRun. Mandatory for importing.
        :param metrics:         The model's metrics.
        :param extra_data:      Extra artifacts and files to log with the model.

        :return: The newly imported ModelArtifact.
        """
        # Validate key is given:
        if key is None:
            raise ValueError(
                "MLFlow models require a key to import into MLRun - the key of the model artifact to be created."
            )

        # Setup defaults:
        metrics = metrics or {}
        extra_data = extra_data or {}
        with tempfile.TemporaryDirectory() as temp_dir:
            # Log the model:
            model = self._log_model(
                project=project,
                model_uri=reference_id,
                key=key,
                metrics=metrics,
                extra_data=extra_data,
                tmp_path=temp_dir,
            )

            logger.info("model imported successfully", key=key)
            return model

    def import_artifact(
        self, project: MlrunProject, reference_id: str, key: Optional[str] = None
    ) -> Artifact:
        """
        Import an artifact from MLFlow to MLRun.

        :param project:      The MLRun project to import the artifact to.
        :param reference_id: The MLFlow artifact uri to import.
        :param key:          The artifact key to be used in MLRun. Mandatory for importing.

        :return: The newly imported artifact.
        """
        # Validate key is given:
        if key is None:
            raise ValueError(
                "MLFlow artifacts require a key to import into MLRun - the key of the artifact to be created."
            )

        # Import the artifact:
        with tempfile.TemporaryDirectory() as tmp_dir:
            # Download the artifact to local temp:
            local_path = mlflow.artifacts.download_artifacts(
                artifact_uri=reference_id, dst_path=tmp_dir
            )

            # Log and return the artifact:
            artifact = self._log_artifact(
                project=project,
                key=key,
                local_path=local_path,
                tmp_path=tmp_dir,
            )

            logger.info("Artifact imported successfully", key=key)
            return artifact

    @staticmethod
    def _log_run(
        context: MLClientCtx, run: mlflow.entities.Run, is_offline: bool = False
    ):
        """
        Log the given MLFlow run to MLRun.

        :param context:     Current MLRun context or project.
        :param run:         MLFlow Run to log. Can be given as a `Run` object
        :param is_offline:  True if logging an offline run (importing), False if online run (tracking)
        """
        client = mlflow.MlflowClient()

        # Create a temporary directory to log all data temporarily:
        with tempfile.TemporaryDirectory() as tmp_dir:
            # When running in offline mode (manually importing) we want to make sure there is an `artifact_path` set in
            # the context, if not we will set it to this temporary directory so the user won't have garbage in his
            # local work dir:
            if is_offline and not context.artifact_path:
                context.artifact_path = tmp_dir

            # Set new MLRun related tags to the MLFlow run:
            client.set_tag(
                run_id=run.info.run_id, key="mlrun-context-id", value=context.uid
            )
            client.set_tag(
                run_id=run.info.run_id, key="mlrun-project-name", value=context.project
            )

            # Get the MLFlow run's tags and save them as labels:
            context.set_label(key="mlflow-user", value=run.data.tags.get("mlflow.user"))
            context.set_label(
                key="mlflow-run-name", value=run.data.tags.get("mlflow.runName")
            )
            context.set_label(key="mlflow-run-id", value=run.info.run_id)
            context.set_label(key="mlflow-experiment-id", value=run.info.experiment_id)

            # Get the MLFlow run's parameters and save as job parameters:
            context._parameters.update(run.data.params)

            # Get the MLFlow run's metrics and log them as results:
            results = run.data.metrics
            context.log_results(results=results)

            # Import the MLFlow run's artifacts to MLRun (model are logged after the rest of artifacts
            # so the artifacts can be registered as extra data in the models):
            artifacts = {}
            model_paths = []
            for artifact in client.list_artifacts(run_id=run.info.run_id):
                # Get the artifact's local path (MLFlow suggests that if the artifact is already in the local filesystem
                # its local path will be returned:
                artifact_local_path = mlflow.artifacts.download_artifacts(
                    run_id=run.info.run_id,
                    artifact_path=artifact.path,
                )
                # Check if the artifact is a model (will be logged after the artifacts):
                if artifact.is_dir and os.path.exists(
                    os.path.join(
                        artifact_local_path, "MLmodel"
                    )  # Add tag to show model dir
                ):
                    model_paths.append(artifact_local_path)
                else:
                    # Log the artifact:
                    artifact = MLFlowTracker._log_artifact(
                        context=context,
                        key=pathlib.Path(artifact.path).name.replace(".", "_"),
                        # Mlflow has the same name for files but with different extensions, so we add extension to name
                        local_path=artifact_local_path,
                        tmp_path=tmp_dir,
                    )
                    artifacts[artifact.key] = artifact

            for model_path in model_paths:
                MLFlowTracker._log_model(
                    context=context,
                    model_uri=model_path,
                    key=pathlib.Path(model_path).stem,
                    metrics=results,
                    extra_data=artifacts,
                    tmp_path=tmp_dir,
                )

    @staticmethod
    def _log_model(
        model_uri: str,
        key: str,
        metrics: dict,
        extra_data: dict,
        tmp_path: os.path,
        context: MLClientCtx = None,
        project: MlrunProject = None,
    ):
        """
        Log the given produced model from MLFlow as a model artifact in MLRun.

        :param model_uri:   The local path to the model (an MLFlow model directory locally downloaded).
        :param key:         The model artifact's key.
        :param metrics:     The key/value dict of model metrics
        :param extra_data:  The extra data to log in addition to the model (training data for example)
        :param tmp_path:    The path to the dir where we temporarily save model and artifacts
        :param context:     The MLRun context to log to, needed one of context or project.
        :param project:     The MLRun project to log to, needed one of context or project.
        """
        # Check that either project or context is provided:
        if not project and not context:
            logger.error(
                "One of context or project must be given in order to log model"
            )
            return

        # Get the model info from MLFlow:
        model_info = mlflow.models.get_model_info(model_uri=model_uri)

        # Prepare the archive path:
        model_uri = pathlib.Path(model_uri)
        archive_path = pathlib.Path(tmp_path) / f"{model_uri.stem}.zip"
        if not os.path.exists(model_uri):
            local_path = mlflow.artifacts.download_artifacts(
                artifact_uri=str(model_uri)
            )
            model_uri = pathlib.Path(local_path)

        # TODO add progress bar for the case of large files
        # Zip the artifact:
        with zipfile.ZipFile(archive_path, "w") as zip_file:
            for path in model_uri.rglob("*"):
                zip_file.write(filename=path, arcname=path.relative_to(model_uri))

        # Get inputs and outputs info:
        inputs = outputs = None
        if model_info.signature is not None:
            if model_info.signature.inputs is not None:
                inputs = MLFlowTracker._schema_to_feature(
                    schema=model_info.signature.inputs
                )
            if model_info.signature.outputs is not None:
                outputs = MLFlowTracker._schema_to_feature(
                    schema=model_info.signature.outputs
                )
        # Log the model:

        kwargs = {
            "key": key,
            "framework": "mlflow",
            "model_file": str(archive_path),
            "metrics": metrics,
            "labels": {
                "mlflow_flavors": str(list(model_info.flavors.keys())),
                "mlflow_run_id": model_info.run_id,
                "mlflow_version": model_info.mlflow_version,
                "mlflow_model_uuid": model_info.model_uuid,
            },
            "extra_data": extra_data,
            "inputs": inputs,
            "outputs": outputs,
        }
        return context.log_model(**kwargs) if context else project.log_model(**kwargs)

    @staticmethod
    def _log_artifact(
        key: str,
        local_path: str,
        tmp_path: str,
        context: MLClientCtx = None,
        project: MlrunProject = None,
    ) -> Artifact:
        """
        Log the given produced file from MLFlow as a run artifact in MLRun.

        :param key:         The artifact's key.
        :param local_path:  The local path to the artifact.
        :param tmp_path:    The path to the dir where we temporarily save artifacts
        :param context:     The MLRun context to log to, needed one of context or project.
        :param project:     The MLRun project to log to, needed one of context or project.
        """
        # Check that either project or context is provided:
        if not project and not context:
            logger.error(
                "One of context or project must be given in order to log artifact"
            )
            return
        # Check if the artifact is a directory for archiving it:
        if pathlib.Path(local_path).is_dir():
            # Prepare the archive path:
            archive_path = pathlib.Path(tmp_path) / f"{key}.zip"
            local_path = pathlib.Path(local_path)
            # Zip the artifact:
            with zipfile.ZipFile(archive_path, "w") as zip_file:
                for path in local_path.rglob("*"):
                    zip_file.write(filename=path, arcname=path.relative_to(local_path))
            # Set the local path to the archive file:
            local_path = str(archive_path)
        kwargs = {
            "item": key,
            "local_path": local_path,
        }
        # Log and return the artifact in the local path:
        return (
            context.log_artifact(**kwargs)
            if context
            else project.log_artifact(**kwargs)
        )

    @staticmethod
    def _schema_to_feature(schema: mlflow.types.Schema) -> list[Feature]:
        """
        Cast MLFlow schema to MLRun features.

        :param schema: The MLFlow schema.

        :return: List of MLRun features representing the schema.
        """
        # Import here due to circular dependencies:
        from mlrun.frameworks._common import CommonUtils

        is_tensor = schema.is_tensor_spec()

        features = []

        for i, item in enumerate(schema.inputs):
            name = item.name or str(i)
            shape = None
            if is_tensor:
                value_type = item.type
                shape = list(item.shape) if item.shape else None
            else:
                value_type = item.type.to_numpy()
            features.append(
                Feature(
                    CommonUtils.convert_np_dtype_to_value_type(value_type),
                    shape,
                    name=name,
                )
            )

        return features

    @staticmethod
    def _create_run_object(handler, run_name, project_name, uid):
        """
        A util function for creating a RunObject
        :param handler:         The handler for MLRun's RunObject
        :param run_name:        The name of the run
        :param project_name:    The name of the project
        :param uid:             The uid of the run
        :return: RunObject
        """
        task = mlrun.new_task(handler=handler, name=run_name, project=project_name)
        task.metadata.uid = uid
        return RunObject.from_template(task)



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import tempfile
from collections.abc import Iterator
from copy import deepcopy
from importlib import import_module
from typing import Optional, Union

import mlrun
import mlrun.artifacts
from mlrun.artifacts import Artifact, ArtifactSpec
from mlrun.model import ModelObj

from ..utils import generate_artifact_uri
from .base import ArtifactStatus


class DocumentLoaderSpec(ModelObj):
    """
    A class to load a document from a file path using a specified loader class.

    This class is responsible for loading documents from a given source path using a specified loader class.
    The loader class is dynamically imported and instantiated with the provided arguments. The loaded documents
    can be optionally uploaded as artifacts. Note that only loader classes that return single results
    (e.g., TextLoader, UnstructuredHTMLLoader, WebBaseLoader(scalar)) are supported - loaders returning multiple
    results like DirectoryLoader or WebBaseLoader(list) are not compatible.

    Attributes:
        loader_class_name (str): The name of the loader class to use for loading documents.
        src_name (str): The name of the source attribute to pass to the loader class.
        kwargs (Optional[dict]): Additional keyword arguments to pass to the loader class.

    """

    _dict_fields = ["loader_class_name", "src_name", "download_object", "kwargs"]

    def __init__(
        self,
        loader_class_name: str = "langchain_community.document_loaders.TextLoader",
        src_name: str = "file_path",
        download_object: bool = True,
        kwargs: Optional[dict] = None,
    ):
        """
        Initialize the document loader.

        Args:
            loader_class_name (str): The name of the loader class to use.
            src_name (str): The source name for the document.
            kwargs (Optional[dict]): Additional keyword arguments to pass to the loader class.
            download_object (bool, optional): If True, the file will be downloaded before launching
                the loader. If False, the loader accepts a link that should not be downloaded.
                Defaults to True.
        Example:
            >>> # Create a loader specification for PDF documents
            >>> loader_spec = DocumentLoaderSpec(
            ...     loader_class_name="langchain_community.document_loaders.PDFLoader",
            ...     src_name="file_path",
            ...     kwargs={"extract_images": True},
            ... )
            >>> # Create a loader instance for a specific PDF file
            >>> pdf_loader = loader_spec.make_loader("/path/to/document.pdf")
            >>> # Load the documents
            >>> documents = pdf_loader.load()

        """
        self.loader_class_name = loader_class_name
        self.src_name = src_name
        self.download_object = download_object
        self.kwargs = kwargs

    def make_loader(self, src_path):
        module_name, class_name = self.loader_class_name.rsplit(".", 1)
        module = import_module(module_name)
        loader_class = getattr(module, class_name)
        kwargs = deepcopy(self.kwargs or {})
        kwargs[self.src_name] = src_path
        loader = loader_class(**kwargs)
        return loader


class MLRunLoader:
    """
    A factory class for creating instances of a dynamically defined document loader.

    Args:
        artifact_key (str, optional): The key for the artifact to be logged.
            The '%%' pattern in the key will be replaced by the source path
            with any unsupported characters converted to '_'. Defaults to "%%".
        local_path (str): The source path of the document to be loaded.
        loader_spec (DocumentLoaderSpec): Specification for the document loader.
        producer (Optional[Union[MlrunProject, str, MLClientCtx]], optional): The producer of the document.
                                If not specified, will try to get the current MLRun context or project.
                                Defaults to None.
        upload (bool, optional): Flag indicating whether to upload the document.
        labels (Optional[Dict[str, str]], optional): Key-value labels to attach to the artifact. Defaults to None.
        tag (str, optional): Version tag for the artifact. Defaults to "".

    Returns:
        DynamicDocumentLoader: An instance of a dynamically defined subclass of BaseLoader.

    Example:
        >>> # Create a document loader specification
        >>> loader_spec = DocumentLoaderSpec(
        ...     loader_class_name="langchain_community.document_loaders.TextLoader",
        ...     src_name="file_path",
        ... )
        >>> # Create a basic loader for a single file
        >>> loader = MLRunLoader(
        ...     source_path="/path/to/document.txt",
        ...     loader_spec=loader_spec,
        ...     artifact_key="my_doc",
        ...     producer=project,
        ...     upload=True,
        ... )
        >>> documents = loader.load()
        >>> # Create a loader with auto-generated keys
        >>> loader = MLRunLoader(
        ...     source_path="/path/to/document.txt",
        ...     loader_spec=loader_spec,
        ...     artifact_key="%%",  # %% will be replaced with encoded path
        ...     producer=project,
        ... )
        >>> documents = loader.load()
        >>> # Use with DirectoryLoader
        >>> from langchain_community.document_loaders import DirectoryLoader
        >>> dir_loader = DirectoryLoader(
        ...     "/path/to/directory",
        ...     glob="**/*.txt",
        ...     loader_cls=MLRunLoader,
        ...     loader_kwargs={
        ...         "loader_spec": loader_spec,
        ...         "artifact_key": "%%",
        ...         "producer": project,
        ...         "upload": True,
        ...     },
        ... )
        >>> documents = dir_loader.load()

    """

    def __new__(
        cls,
        source_path: str,
        loader_spec: "DocumentLoaderSpec",
        artifact_key="%%",
        producer: Optional[Union["MlrunProject", str, "MLClientCtx"]] = None,  # noqa: F821
        upload: bool = False,
        tag: str = "",
        labels: Optional[dict[str, str]] = None,
    ):
        # Dynamically import BaseLoader
        from langchain_community.document_loaders.base import BaseLoader

        class DynamicDocumentLoader(BaseLoader):
            def __init__(
                self,
                local_path,
                loader_spec,
                artifact_key,
                producer,
                upload,
                tag,
                labels,
            ):
                self.producer = producer
                self.artifact_key = (
                    MLRunLoader.artifact_key_instance(artifact_key, local_path)
                    if "%%" in artifact_key
                    else artifact_key
                )
                self.loader_spec = loader_spec
                self.local_path = local_path
                self.upload = upload
                self.tag = tag
                self.labels = labels

                # Resolve the producer
                if not self.producer:
                    self.producer = mlrun.mlconf.active_project
                if isinstance(self.producer, str):
                    self.producer = mlrun.get_or_create_project(self.producer)

            def lazy_load(self) -> Iterator["Document"]:  # noqa: F821
                collections = None
                try:
                    artifact = self.producer.get_artifact(self.artifact_key, self.tag)
                    collections = (
                        artifact.status.collections if artifact else collections
                    )
                except mlrun.MLRunNotFoundError:
                    pass
                artifact = self.producer.log_document(
                    key=self.artifact_key,
                    document_loader_spec=self.loader_spec,
                    local_path=self.local_path,
                    upload=self.upload,
                    labels=self.labels,
                    tag=self.tag,
                    collections=collections,
                )
                res = artifact.to_langchain_documents()
                return res

        # Return an instance of the dynamically defined subclass
        instance = DynamicDocumentLoader(
            artifact_key=artifact_key,
            local_path=source_path,
            loader_spec=loader_spec,
            producer=producer,
            upload=upload,
            tag=tag,
            labels=labels,
        )
        return instance

    @staticmethod
    def artifact_key_instance(artifact_key: str, src_path: str) -> str:
        if "%%" in artifact_key:
            resolved_path = DocumentArtifact.key_from_source(src_path)
            artifact_key = artifact_key.replace("%%", resolved_path)
        return artifact_key


class DocumentArtifact(Artifact):
    """
    A specific artifact class inheriting from generic artifact, used to maintain Document meta-data.
    """

    @staticmethod
    def key_from_source(src_path: str) -> str:
        """Convert a source path into a valid artifact key by replacing invalid characters with underscores.
        Args:
            src_path (str): The source path to be converted into a valid artifact key
        Returns:
            str: A modified version of the source path where all invalid characters are replaced
                with underscores while preserving valid sequences in their original positions
        Examples:
            >>> DocumentArtifact.key_from_source("data/file-name(v1).txt")
            "data_file-name_v1__txt"
        """
        pattern = mlrun.utils.regex.artifact_key[0]
        # Convert anchored pattern (^...$) to non-anchored version for finditer
        search_pattern = pattern.strip("^$")
        result = []
        current_pos = 0

        # Find all valid sequences
        for match in re.finditer(search_pattern, src_path):
            # Add '_' values for characters between matches
            for char in src_path[current_pos : match.start()]:
                result.append("_")

            # Add the valid sequence
            result.append(match.group())
            current_pos = match.end()

        # Handle any remaining characters after the last match
        for char in src_path[current_pos:]:
            result.append("_")

        resolved_path = "".join(result)
        resolved_path = resolved_path.lstrip("_")
        return resolved_path

    class DocumentArtifactSpec(ArtifactSpec):
        _dict_fields = ArtifactSpec._dict_fields + [
            "document_loader",
            "original_source",
        ]

        def __init__(
            self,
            *args,
            document_loader: Optional[DocumentLoaderSpec] = None,
            original_source: Optional[str] = None,
            **kwargs,
        ):
            super().__init__(*args, **kwargs)
            self.document_loader = document_loader
            self.original_source = original_source

    class DocumentArtifactStatus(ArtifactStatus):
        _dict_fields = ArtifactStatus._dict_fields + ["collections"]

        def __init__(
            self,
            *args,
            collections: Optional[dict] = None,
            **kwargs,
        ):
            super().__init__(*args, **kwargs)
            self.collections = collections if collections is not None else {}

    kind = "document"

    METADATA_SOURCE_KEY = "source"
    METADATA_ORIGINAL_SOURCE_KEY = "original_source"
    METADATA_CHUNK_KEY = "mlrun_chunk"
    METADATA_ARTIFACT_TARGET_PATH_KEY = "mlrun_target_path"
    METADATA_ARTIFACT_TAG = "mlrun_tag"
    METADATA_ARTIFACT_KEY = "mlrun_key"
    METADATA_ARTIFACT_PROJECT = "mlrun_project"

    def __init__(
        self,
        original_source: Optional[str] = None,
        document_loader_spec: Optional[DocumentLoaderSpec] = None,
        collections: Optional[dict] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.spec.document_loader = (
            document_loader_spec.to_dict()
            if document_loader_spec
            else self.spec.document_loader
        )
        self.spec.original_source = original_source or self.spec.original_source
        self.status = DocumentArtifact.DocumentArtifactStatus(collections=collections)

    @property
    def status(self) -> DocumentArtifactStatus:
        return self._status

    @status.setter
    def status(self, status):
        self._status = self._verify_dict(
            status, "status", DocumentArtifact.DocumentArtifactStatus
        )

    @property
    def spec(self) -> DocumentArtifactSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(
            spec, "spec", DocumentArtifact.DocumentArtifactSpec
        )

    def get_source(self):
        """Get the source URI for this artifact."""
        return generate_artifact_uri(self.metadata.project, self.spec.db_key)

    def to_langchain_documents(
        self,
        splitter: Optional["TextSplitter"] = None,  # noqa: F821
    ) -> list["Document"]:  # noqa: F821
        # Try new langchain 1.0+ import path first
        try:
            from langchain_core.documents import Document
        except ImportError:
            # Fall back to old langchain <1.0 import path
            from langchain.schema import Document

        """
        Create LC documents from the artifact

        Args:
            splitter (Optional[TextSplitter]): A LangChain TextSplitter to split the document into chunks.

        Returns:
            list[Document]: A list of LangChain Document objects.
        """

        loader_spec = DocumentLoaderSpec.from_dict(self.spec.document_loader)
        if loader_spec.download_object and self.get_target_path():
            with tempfile.NamedTemporaryFile() as tmp_file:
                mlrun.datastore.store_manager.object(
                    url=self.get_target_path()
                ).download(tmp_file.name)
                loader = loader_spec.make_loader(tmp_file.name)
                documents = loader.load()
        elif self.spec.original_source:
            loader = loader_spec.make_loader(self.spec.original_source)
            documents = loader.load()
        else:
            raise ValueError(
                "No src_path or target_path provided. Cannot load document."
            )

        results = []
        idx = 0
        for document in documents:
            if splitter:
                texts = splitter.split_text(document.page_content)
            else:
                texts = [document.page_content]

            metadata = document.metadata

            metadata[self.METADATA_ORIGINAL_SOURCE_KEY] = self.spec.original_source
            metadata[self.METADATA_SOURCE_KEY] = self.get_source()
            metadata[self.METADATA_ARTIFACT_TAG] = self.tag or "latest"
            metadata[self.METADATA_ARTIFACT_KEY] = self.db_key
            metadata[self.METADATA_ARTIFACT_PROJECT] = self.metadata.project

            if self.get_target_path():
                metadata[self.METADATA_ARTIFACT_TARGET_PATH_KEY] = (
                    self.get_target_path()
                )

            for text in texts:
                metadata[self.METADATA_CHUNK_KEY] = str(idx)
                doc = Document(
                    page_content=text,
                    metadata=metadata.copy(),
                )
                results.append(doc)
                idx = idx + 1
        return results

    def collection_add(self, collection_id: str) -> bool:
        """
        Add a collection ID to the artifact's collection list.

        Adds the specified collection ID to the artifact's collection mapping if it
        doesn't already exist.
        This method only modifies the client-side artifact object and does not persist
        the changes to the MLRun DB. To save the changes permanently, you must call
        project.update_artifact() after this method.

        Args:
            collection_id (str): The ID of the collection to add
        """
        if collection_id not in self.status.collections:
            self.status.collections[collection_id] = "1"
            return True
        return False

    def collection_remove(self, collection_id: str) -> bool:
        """
        Remove a collection ID from the artifact's collection list.

        Removes the specified collection ID from the artifact's local collection mapping.
        This method only modifies the client-side artifact object and does not persist
        the changes to the MLRun DB. To save the changes permanently, you must call
        project.update_artifact() or context.update_artifact() after this method.

        Args:
            collection_id (str): The ID of the collection to remove
        """
        if collection_id in self.status.collections:
            self.status.collections.pop(collection_id)
            return True
        return False
